class AlvPossibleFormatError(Exception):
    '''
    Cannot recognize the input format
    '''
    def __init__(self, expression):
        self.expression = 'HUBBA'
        self.message = 'File format not recognized.'

class AlvEmptyAlignment(Exception):
    def __init__(self):
        self.message = 'Input alignment contains no sequences.'



from .alignment import AlignmentBlock
import random
import sys
import shutil



class AlignmentTerminal:
    '''
    This class encapsulates knowledge about the terminal and how to draw on it.
    '''
    def __init__(self, terminal_width, random_sample_size=0, sorting=None, order=None, accession_pattern=False):
        self.width = terminal_width
        self.random_sample_size = random_sample_size
        self.sorting = sorting
        self.order = order
        self.selection = accession_pattern


    def get_accession_list(self, al):
        '''
        Choose accessions and their order based on the user-choices.
        '''
        if self.sorting == 'alpha':
            accessions = al.sorted_accessions()
        elif self.sorting == 'fixed':
            accessions = self.order
        elif self.sorting == 'by identity':
            accessions = al.sort_by_identity(self.order)
        else:
            accessions = al.accessions()

        if self.selection:
            chosen_accessions = []
            for acc in accessions:
                if self.selection in acc:
                    chosen_accessions.append(acc)
            return chosen_accessions
        else:
            return list(accessions)


    def _output_block(self, al, painter, width, chosen_accessions, block, dotted):
        '''
        Write one block of an aligmnent to the terminal.
        The accessions to the left, sequences to the right, and at the bottom a tick string.
        '''
        if dotted:
            template_acc=chosen_accessions[0]
            chosen_accessions = chosen_accessions[1:]
            colored_subseq = al.apply_painter(template_acc, block, painter)
            print("{0:{width}}{1}".format(template_acc, colored_subseq, width=self.left_margin))
            
        for acc in chosen_accessions:
            if dotted:
                colored_subseq = al.apply_dotter(acc, block, painter, template_acc)
            else:
                colored_subseq = al.apply_painter(acc, block, painter)
            print("{0:{width}}{1}".format(acc, colored_subseq, width=self.left_margin))
        print(make_tick_string(self.left_margin, block.start, block.end, 20, 7))


    def _setup_left_margin(self, al, chosen_accessions):
        '''
        Compute the width of the left margin from the accession lengths
        '''
        self.left_margin = 1 + al.accession_widths(chosen_accessions)
        assert self.left_margin < self.width - 10


    def output_alignment(self, al, painter, chosen_width, dotted=False):
        '''
        Output alignment al to stdout in blocks of width at most w with colors from painter.

        Args:
          al -- the alignment object
          painter -- Painter object that decides colors for symbols
          chosen_width -- How many columns to use for one alignment block
          dotted -- Flag. Output periods (.) if position identical to first sequence?
        '''
        chosen_accessions =  self.get_accession_list(al)

        if self.random_sample_size and len(chosen_accessions) > self.random_sample_size:
            chosen_accessions = random.sample(chosen_accessions, self.random_sample_size)

        self._setup_left_margin(al, chosen_accessions)

        columns_per_block = al.block_width(self.width, chosen_width)
        for block in al.blocks(columns_per_block):
            self._output_block(al, painter, chosen_width, chosen_accessions, block, dotted)


    def output_glimpse(self, al, painter, chosen_width, dotted=False):
        '''
        Output a single-screen glimpse of the alignment. An attempt at finding
        the most interesting (guessed to be the most conserved part of a random
        sample of sequences) of the alignment is done.
        '''
        chosen_accessions =  self.get_accession_list(al)

        if self.height:
            n_seqs_to_view = min(len(chosen_accessions), self.height - 2) # -2 to make room for tick line and next prompt on terminals
        else:
            n_seqs_to_view = min(len(chosen_accessions), 20)
        chosen_accessions = random.sample(chosen_accessions, n_seqs_to_view)

        self._setup_left_margin(al, chosen_accessions)

        n_columns = al.block_width(self.width, chosen_width) # This many alignment columns
        conserved_block = al.get_conserved_block(n_columns)

        self._output_block(al, painter, chosen_width, chosen_accessions, conserved_block, dotted)


def calc_tick_indices(start, end, distance, min_distance):
    '''
    Return a list of indices for which we want a tick mark at the bottom of the alignment.
    The goal is to have an index for the starting position of a block (leftmost column number),
    and then a tick mark on even multiples of 20 (or what is given by 'distance'), for example:
        53   60                  80                 100
    Care is needed so that space is left between first and second indices, and min_distance indicates
    how much.
    '''
    first_even_pos = (start // distance + 1) * distance
    if first_even_pos - start < min_distance:
        first_even_pos += distance # Compensate a bit
    positions = range(first_even_pos, end, distance)
    return positions

def make_one_tick(position, space, tickmark):
    '''
    Return a string which is 'space' wide and contains a number (the position)
    followed by an up-arrow.
    '''
    return '{0:>{width}}{tickmark}'.format(position, width=space-1,tickmark=tickmark)

def make_tick_string(left_margin, start, end, distance, min_distance):
    '''
    Construct the index bar which is printed at the bottom of an alignment block.

    left_margin is how much space is allowed for accessions.
    start is the column number of the beginning of an alignment block.
    end is the last column of an alignment block.
    distance is the desired distance between up-arrows
    min_distance is the space we allow for position numbers plus an up-arrow
    '''
    even_indices = calc_tick_indices(start, end, distance, min_distance)

    tickmark = '^'
    if sys.stdout.encoding == 'UTF-8':
        tickmark = '↑'

    # Initial space
    index_bar = ' ' * (left_margin - min_distance + 1) # Account for space needed by indices

    # Add first column index
    index_bar += make_one_tick(start, min(left_margin+1, min_distance), tickmark)

    last_pos = start
    for pos in even_indices:
        spacer = pos - last_pos - 1
        index_block = '{0:>{width}}{tickmark}'.format(pos, width=spacer, tickmark=tickmark)
        index_bar += index_block
        last_pos = pos
    return index_bar


class AlignmentShellTerminal(AlignmentTerminal):
    '''
    This subclass gets the attributes from an argparse object and
    checking the actual terminal size.
    '''
    def __init__(self, args):
        self.random_sample_size = args.random_accessions

        sorting = args.sorting
        order = None
        if args.sort_by_id:
            sorting = 'by identity'
            order = args.sort_by_id
        elif args.sorting_order:
            sorting = 'fixed'
            order = args.sorting_order.split(',')
            if len(self.order) == 0:
                raise Exception('Bad order specification: no accessions in input')
        if args.select_matching:
            accession_pattern = args.select_matching
        else:
            accession_pattern = False

        terminal_width, terminal_height = shutil.get_terminal_size()
        super().__init__(terminal_width, args.random_accessions, sorting, order, accession_pattern)
        self.height = terminal_height





import Bio.Seq
import itertools
import math
import random
from collections import Counter
from .exceptions import AlvEmptyAlignment

class BaseAlignment:
    '''
    Base class for alignments. It contains the interface to use
    to access all alignments, and implementations for the subclasses
    with single-letter column widths.
    '''
    def __init__(self, alignment):
        self.al = alignment   # Holder of the BioPython alignment object
        self.al_length = alignment.get_alignment_length()
        self.type = None
        self.column_width = 1
        self._update_seq_index()
        self.columns = self._summarize_columns()
        self.basic_info = {'Number of sequences': len(self.al),
                           'Alignment width': self.al_length}
        self.indels=['.', '-', ' ']

    def _update_seq_index(self):
        self.seq_indices = { r.id : i for i, r in enumerate(self.al)} # Get a dictionary mapping accession to row index in alignment

    def al_width(self):
        '''
        The number of columns in the alignment.
        '''
        return self.al_length

    def trim_accessions(self, start, stop):
        for record in self.al:
            acc = record.id
            new_acc = acc[start:stop]
            record.id = new_acc
        self._update_seq_index()

    def abbreviate_accessions(self, n_chars):
        for record in self.al:
            acc = record.id
            if len(acc) > n_chars:
                new_acc = acc[0:n_chars] + '*' + acc[-n_chars:]
                record.id = new_acc
        self._update_seq_index()

    def accessions(self):
        '''
        The accessions in arbitrary order?
        '''
        return map(lambda r: r.id, self.al)

    def sorted_accessions(self):
        '''
        Accessions in alphabetical order.
        '''
        return sorted(self.accessions())

    def sort_by_identity(self, acc):
        '''
        Return accessions in order by similarity (as percent identity) to sequence 'acc'.
        '''
        if acc not in self.seq_indices:
            raise ValueError('Accession "' + acc + '" not found in the alignment')
        pivot_sequence_record = self.al[self.seq_indices[acc]]
        sorted_accessions = map(lambda rec: rec.id,
                                sorted(self.al,
                                       key=lambda rec: percent_identity(pivot_sequence_record.seq, rec),
                                       reverse=True))
        return sorted_accessions

    def random_accessions(self, n):
        '''
        Return a random sample of accessions, sample size = n.
        '''
        return random.choice(list(self.accessions()))

    def accession_widths(self, accessions=None):
        '''
        Compute the space needed for all accessions, plus one for
        a delimiter.
        '''
        max_accession_length = 5        # initial guess, or minimum
        for record in self.al:
            if (not accessions or record.id in accessions) and len(record.id) > max_accession_length:
                max_accession_length = len(record.id)
        return max_accession_length


    def block_width(self, terminal_width, args_width):
        '''
        For wide alignments, we need to break it up in blocks.
        This method calculates how many characters to output in a block.

        Take the margin size (for accessions) into account and avoid ending up
        with blocks of size 1.
        '''
        if args_width == 0:
            al_width = self.al_length
            left_margin = 1 + self.accession_widths() # Add 1 for a space to the right of the accessions
            return self._compute_block_width(terminal_width, al_width, left_margin)
        else:
            return args_width

    def _compute_block_width(self, terminal_width, al_width, left_margin):
        '''
        Helper for block_width().
        '''
        last_block_width = al_width % (terminal_width - left_margin)
        n_blocks = al_width // (terminal_width - left_margin) # Not counting last block, if uneven

        if last_block_width == al_width:
            # Short alignment, fits in one block
            return al_width
        elif last_block_width > 10:
            return terminal_width - left_margin
        else:
            sacrifice = max([1, (10 - last_block_width) // n_blocks])
            return terminal_width - left_margin - sacrifice

    def blocks(self, block_width):
        al_width = self.al_length
        if al_width == 0:
            raise AlvEmptyAlignment()
        else:
            for start in range(0, al_width, block_width):
                end = min(al_width, start + block_width)
                yield AlignmentBlock(start, end)


    def apply_painter(self, acc, block, painter):
        '''
        Colorize a subsequence according to some style (painter).
        This implementation works for aa and plain DNA/RNA.
        '''
        idx = self.seq_indices[acc]
        seq_record = self.al[idx, block.start:block.end]
        colored_seq = ''
        for col_no, c in enumerate(seq_record.seq):
            colored_seq += painter.colorizer(c, self.columns[block.start + col_no])
        return painter.sol() + colored_seq + painter.eol()

    def apply_dotter(self, acc, block, painter, template_acc):
        '''
        Colorize and adapt a subsequence according to some style (painter) and
        following accession for a template sequence. Write '.' if the site is conserved.
        This implementation works for aa and plain DNA/RNA.
        '''
        idx = self.seq_indices[acc]
        seq_record = self.al[idx, block.start:block.end]

        template_idx = self.seq_indices[template_acc]
        template_record = self.al[template_idx, block.start:block.end]
        template_seq = template_record.seq

        colored_seq = ''
        for col_no, c in enumerate(seq_record.seq):
            if c == template_seq[col_no]:
                colored_seq += '.'
            else:
                colored_seq += painter.colorizer(c, self.columns[block.start + col_no])
        return painter.sol() + colored_seq + painter.eol()

    def _summarize_columns(self):
        '''
        Count the different elements in each column.
        '''
        columns = []
        for col_no in range(self.al_length):
            columns.append(Counter(self.al[:, col_no]))
        return columns

    def get_basic_info(self):
        '''
        Generator for pairs of description and value, e.g., "alignment width" and an integer.
        '''
        self.basic_info['Sequence type'] = self.type
        for descr, val in self.basic_info.items():
            yield descr, val

    def get_column_conservation(self):
        '''
        Return a list of floats representing how conserved the columns are.
        In this first version of the method, conservation equals fraction of elements
        identical to the majority element.
        '''
        result = []       # For the return value
        column_summaries = self._summarize_columns()
        for c in column_summaries:
            for indel in self.indels:
                del c[indel]
            if len(c) == 0:
                result.append(0)
            else:
                majority = c.most_common(1)[0][1]
                conservation = majority / len(self.al)
                result.append(conservation)
        return result


    def get_conserved_block(self, n_columns):
        '''
        Return a block that is a suitable representation of the alignment.
        Used for alignment glimpses.
        '''
        conservation = self.get_column_conservation()   # A list of conservation scores. Want a maximised "window"
        accumulated_conservation = []
        acc = 0
        for c in conservation:
            acc += c
            accumulated_conservation.append(acc)

        if self.al_width() <= n_columns:
            return AlignmentBlock(0, self.al_width())
        else:
            best_start = max(range(self.al_width() - n_columns),
                             key=lambda i: accumulated_conservation[i+n_columns] - accumulated_conservation[i])
            return AlignmentBlock(best_start, best_start + n_columns)



class AminoAcidAlignment(BaseAlignment):
    def __init__(self, alignment):
       	super().__init__(alignment)
        self.type = 'aa'

class DnaAlignment(BaseAlignment):
    def __init__(self, alignment):
       	super().__init__(alignment)
        self.type = 'dna'

class CodonAlignment(BaseAlignment):
    '''
    Alignment of coding DNA. A column has a width of three nucleotides.
    '''
    def __init__(self, alignment):
        self.type = 'codon'
        self.column_width = 3
        self.genetic_code = 1   # The standard code
        super().__init__(alignment)
        self.basic_info['Genetic code'] = self.genetic_code

    def block_width(self, terminal_width, args):
        '''
        Refinement of the superclass' implemention to ensure that all blocks have
        a width that is a multiple of three.
        '''
        nominal_width = super().block_width(terminal_width, args)
        remainder = nominal_width % 3
        return nominal_width - remainder # Safe, because the superclass method guarantees width >= 10.

    def apply_painter(self, acc, block, painter):
        '''
        Colorize a CODON (!) subsequence according to some style (painter).
        This implementation works for codons.
        '''
        idx = self.seq_indices[acc]
        seq_record = self.al[idx, block.start:block.end]
        seq = str(seq_record.seq)
        colored_seq = ''

        for codon_col, pos in enumerate(range(0, len(seq), 3)):
            c = seq[pos:pos+3]
            colored_seq += painter.colorizer(c, self.columns[block.start // 3 + codon_col])
        return painter.sol() + colored_seq + painter.eol()

    def apply_dotter(self, acc, block, painter, template_acc):
        '''
        Colorize and adapt a CODON(!) subsequence according to some style (painter) and
        following accession for a template sequence. Write '.' if the site is conserved.
        This implementation works for codon sequences.
        '''
        idx = self.seq_indices[acc]
        seq_record = self.al[idx, block.start:block.end]
        seq = str(seq_record.seq)

        template_idx = self.seq_indices[template_acc]
        template_record = self.al[template_idx, block.start:block.end]
        template_seq = template_record.seq

        colored_seq = ''
        for codon_col_no, pos in enumerate(range(0, len(seq), 3)):
            c = seq[pos:pos+3]
            if c == template_seq[pos:pos+3]:
                colored_seq += '...'
            else:
                colored_seq += painter.colorizer(c, self.columns[block.start // 3 + codon_col_no])
        return painter.sol() + colored_seq + painter.eol()



    def _summarize_columns(self):
        '''
        Specialization of base method for codon columns. Do not focus on the amino acids, but look at
        amino acid columns.
        '''
        columns = []
        for pos in range(0, self.al_length, 3):
            codon_column = map(lambda r: str(r.seq), self.al[:, pos:pos+3])
            aa_column = map(lambda codon: self._translate(codon), codon_column)
            columns.append(Counter(aa_column))
        return columns


    def get_conserved_block(self, n_columns):
        '''
        Return a block that is a suitable representation of the alignment.
        Used for alignment glimpses.
        Specialised for codon alignments.
        '''
        conservation = self.get_column_conservation()   # A list of conservation scores. Want a maximised "window"
        accumulated_conservation = []
        acc = 0
        for c in conservation:
            acc += c
            accumulated_conservation.append(acc)

        n_columns = n_columns // 3  # We want codon columns now

        codon_al_width = math.floor(self.al_width() / 3)
        if codon_al_width <= n_columns:
            return AlignmentBlock(0, codon_al_width)
        else:
            best_start = max(range(0, codon_al_width - n_columns),
                             key=lambda i: accumulated_conservation[i+n_columns] - accumulated_conservation[i])
            return AlignmentBlock(3*best_start, 3*(best_start + n_columns))


    def _translate(self, codon):
        try:
            if codon == '---':
                return '-'
            else:
                if len(codon) == 3:
                    aa = Bio.Seq.translate(codon, table=self.genetic_code)
                else:
                    aa = '?'
                return aa
        except Bio.Data.CodonTable.TranslationError:
            # For when we have weird codons/alignments
            return 'X'

    def set_genetic_code(self, code):
        self.genetic_code = code
        self.basic_info['Genetic code'] = code


class AlignmentBlock:
    def __init__(self, start, end):
        self.start = start
        self.end = end

    def __str__(self):
        return f'<AlignmentBlock start={self.start} end={self.end}>'


def percent_identity(seq1, seq2):
    identical = 0
    for c1, c2 in zip(seq1, seq2):
        if c1 == c2:
            identical +=1
    return identical / len(seq1)



import unittest
import argparse
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from Bio.Align import MultipleSeqAlignment
import alv.io
from alv.alignment import AminoAcidAlignment, DnaAlignment, CodonAlignment, BaseAlignment
import alv.alignmentterminal as at
import alv.colorize
import alv.exceptions

class TestIOetc(unittest.TestCase):
    def test_guess_seq_type(self):

        coding = MultipleSeqAlignment([
             SeqRecord(Seq("AAGCTGATCAGC"), annotations={'molecule_type':'DNA'}, id="Alpha"),
             SeqRecord(Seq("CTGAAGATCAGC"), annotations={'molecule_type':'DNA'}, id="Beta"),
             SeqRecord(Seq("AAGCTGATCAGG"), annotations={'molecule_type':'DNA'}, id="Gamma"),
         ])
        noncoding = MultipleSeqAlignment([
             SeqRecord(Seq("CGTCGTCGTCGT"), annotations={'molecule_type':'DNA'}, id="Alpha"),
             SeqRecord(Seq("CGACGACGACGA"), annotations={'molecule_type':'DNA'}, id="Beta"),
             SeqRecord(Seq("ATAATAATAATA"), annotations={'molecule_type':'DNA'}, id="Gamma"),
         ])
        peptide = MultipleSeqAlignment([
             SeqRecord(Seq("ARNDCEQHIKL*"), annotations={'molecule_type':'protein'}, id="Alpha"),
             SeqRecord(Seq("ARNDCEQHIKL*"), annotations={'molecule_type':'protein'}, id="Beta"),
             SeqRecord(Seq("ARNDCEQHIKLM"), annotations={'molecule_type':'protein'}, id="Gamma"),
         ])
        self.assertEqual(alv.io.guess_seq_type(coding), 'codon')
        self.assertEqual(alv.io.guess_seq_type(noncoding), 'dna')
        self.assertEqual(alv.io.guess_seq_type(peptide), 'aa')


class TestFormats(unittest.TestCase):
    def setUp(self):
        self.dna_filename = 'tests/t1.fa'
        self.aa_filename = 'tests/t4.fa'
        self.sthlm_filename= 'tests/t4.sthlm'
        self.pfam_file = 'tests/PF00005_seed.txt'
        self.nexus_filename = 'tests/test.nex'

    def test_reading_fasta_files(self):
        al, painter = alv.io.read_alignment(self.dna_filename, 'dna', 'fasta', '', 'standard')
        self.assertIsInstance(al, DnaAlignment)
        self.assertIsInstance(painter, alv.colorize.DnaPainter)

        al, painter = alv.io.read_alignment(self.dna_filename, 'aa', 'fasta', 'clustal', 'standard')
        self.assertIsInstance(al, AminoAcidAlignment)
        self.assertIsInstance(painter, alv.colorize.AminoAcidPainter)

        al, painter = alv.io.read_alignment(self.dna_filename, 'codon', 'fasta', 'clustal', 'standard')
        self.assertIsInstance(al, CodonAlignment)
        self.assertIsInstance(painter, alv.colorize.CodonPainter)

    def test_reading_stockholm_files(self):
        al, painter = alv.io.read_alignment(self.sthlm_filename, 'aa', 'stockholm', '', 'standard')
        self.assertIsInstance(al, AminoAcidAlignment)
        self.assertIsInstance(painter, alv.colorize.AminoAcidPainter)
        al, painter = alv.io.read_alignment(self.pfam_file, 'aa', 'stockholm', '', 'standard')
        self.assertIsInstance(al, AminoAcidAlignment)
        self.assertIsInstance(painter, alv.colorize.AminoAcidPainter)

    def test_reading_nexus_files(self):
        al, painter = alv.io.read_alignment(self.nexus_filename, 'aa', 'nexus', '', 'standard')
        self.assertIsInstance(al, AminoAcidAlignment)
        self.assertIsInstance(painter, alv.colorize.AminoAcidPainter)

    def test_reading_wrong_format(self):
        with self.assertRaises(ValueError):
            al, painter = alv.io.read_alignment(self.aa_filename, 'aa', 'clustal', '', 'standard')
        with self.assertRaises(ValueError):
            al, painter = alv.io.read_alignment(self.aa_filename, 'aa', 'phylip', '', 'standard')
        with self.assertRaises(ValueError):
            al, painter = alv.io.read_alignment(self.aa_filename, 'aa', 'stockholm', '', 'standard')
        with self.assertRaises(ValueError):
            al, painter = alv.io.read_alignment(self.aa_filename, 'aa', 'nexus', '', 'standard')
        with self.assertRaises(ValueError):
            al, painter = alv.io.read_alignment(self.sthlm_filename, 'aa', 'fasta', '', 'standard')
        with self.assertRaises(ValueError):
            al, painter = alv.io.read_alignment(self.sthlm_filename, 'aa', 'clustal', '', 'standard')
        with self.assertRaises(ValueError):
            al, painter = alv.io.read_alignment(self.sthlm_filename, 'aa', 'phylip', '', 'standard')
        with self.assertRaises(ValueError):
            al, painter = alv.io.read_alignment(self.sthlm_filename, 'aa', 'nexus', '', 'standard')
        with self.assertRaises(ValueError):
            al, painter = alv.io.read_alignment(self.nexus_filename, 'aa', 'fasta', '', 'standard')
        with self.assertRaises(ValueError):
            al, painter = alv.io.read_alignment(self.nexus_filename, 'aa', 'phylip', '', 'standard')
        with self.assertRaises(ValueError):
            al, painter = alv.io.read_alignment(self.nexus_filename, 'aa', 'clustal', '', 'standard')
        with self.assertRaises(ValueError):
            al, painter = alv.io.read_alignment(self.nexus_filename, 'aa', 'stockholm', '', 'standard')


    def test_guessing_seq_type(self):
        '''
        Given simple DNA or AA input, is the sequence typ (DNA/AA) guessed correctly?
        '''
        al, painter = alv.io.read_alignment(self.dna_filename, 'guess', 'fasta', '', 'standard')
        self.assertIsInstance(al, DnaAlignment)
        al, painter = alv.io.read_alignment(self.aa_filename, 'guess', 'fasta', '', 'standard')
        self.assertIsInstance(al, AminoAcidAlignment)
        al, painter = alv.io.read_alignment(self.nexus_filename, 'guess', 'nexus', '', 'standard')
        self.assertIsInstance(al, CodonAlignment)


class TestBlockCalculation(unittest.TestCase):
    def setUp(self):
        self.al1, _ = alv.io.read_alignment('tests/t2.fa', 'dna', 'fasta','', 'standard')
        self.al2, _ = alv.io.read_alignment('tests/t2_longaccessions.fa', 'dna', 'fasta','', 'standard')


    def test_block_width_calculation(self):
        self.assertEqual(10, self.al1.block_width(100, 0))
        self.assertEqual(10, self.al1.block_width(100, 10))
        self.assertEqual(5, self.al1.block_width(100, 5))
        self.assertEqual(10, self.al1.block_width(10, 10))


class TestIndexBar(unittest.TestCase):
    def setUp(self):
        self.tickmark = '^'
        import sys
        if sys.stdout.encoding == 'UTF-8':
            self.tickmark = '↑'

    def test_make_one_tick(self):
        s = at.make_one_tick(5, 10, '^')
        self.assertEqual(len(s), 10)
        self.assertEqual(s, '        5^')

    def test_calc_tick_indices(self):
        indices = list(at.calc_tick_indices(33, 100, 10, 7))
        self.assertEqual(indices, [40, 50, 60, 70, 80, 90])

        indices = list(at.calc_tick_indices(50, 100, 10,10))
        self.assertEqual(indices, [60, 70, 80, 90])

    def test_make_tick_string(self):
        s = at.make_tick_string(6, 0, 40, 40, 10)
        if self.tickmark == '↑':
            self.assertEqual(s, '     0↑')
        else:
            self.assertEqual(s, '     0^')

        s = at.make_tick_string(6, 0, 40, 20, 10)
        if self.tickmark == '↑':
            self.assertEqual(s, '     0↑                 20↑')
        else:
            self.assertEqual(s, '     0^                 20^')

        s = at.make_tick_string(5, 0, 50, 20, 7)
        if self.tickmark == '↑':
            self.assertEqual(s, '    0↑                 20↑                 40↑')
        else:
            self.assertEqual(s, '    0^                 20^                 40^')

        s = at.make_tick_string(7, 35, 100, 20, 10)
        if self.tickmark == '↑':
            self.assertEqual(s, '     35↑                      60↑                 80↑')
        else:
            self.assertEqual(s, '     35^                      60^                 80^')


if __name__ == '__main__':
    unittest.main()



from Bio import AlignIO
from math import log
import sys

from .alignment import AminoAcidAlignment, DnaAlignment, CodonAlignment
from .colorize import AminoAcidPainter, DnaPainter, CodonPainter, AminoAcidTaylorPainter, AminoAcidHydrophobicity
from .exceptions import AlvPossibleFormatError


def guess_format(filename):
    '''
    Returns a string that guesses the MSA format used in filename.
    '''
    with open(filename) as h:
        first_line = h.readline()
        if len(first_line) < 2:
            raise ValueError('Empty file')
        if first_line[:15] == '# STOCKHOLM 1.0':
            return 'stockholm'
        elif first_line == 'CLUSTAL':
            return 'clustal'
        elif first_line[0] == '>':
            return 'fasta'
        elif first_line.lower().startswith("#nexus"):
            return 'nexus'
        else:
            tokens = first_line.split()
            if len(tokens) != 2:
                raise AlvPossibleFormatError(filename) # Don't recognize the format
            else:
                try:
                    _ = int(tokens[0])
                    _ = int(tokens[1])
                except:
                    raise AlvPossibleFormatError(filename)
                # Came this far? Success!
                return 'phylip'


def read_alignment(file, seqtype, input_format, color_scheme, genetic_code, alignment_no=0):
    '''
    Factory function. Read the alignment with BioPython's support, and
    return an appropriate alv alignment.
    '''
    if file == '-':
        file = sys.stdin        # Start reading from stdin if "magic filename"
    alignments = list(AlignIO.parse(file, input_format))

    if alignments:
        n_msas = len(alignments)
        if alignment_no >= n_msas:
            raise IOError(f'Alignment index too large, only {n_msas} alignment(s) in the file.')
        return get_alv_objects(alignments[alignment_no], seqtype, color_scheme, genetic_code)
    else:
        raise ValueError('No alignment in input')

def get_alv_objects(alignment, seqtype, color_scheme, genetic_code):
    '''
    Take the alignment object and return a suitable Alv alignment object, with respect
    to sequence type, and colorization object ("painter").
    '''
    if seqtype == 'guess':
        seqtype = guess_seq_type(alignment)

    if color_scheme == 'taylor':
        painter = AminoAcidTaylorPainter()
    elif color_scheme == 'hydrophobicity':
        painter = AminoAcidHydrophobicity()
    else:
        painter = AminoAcidPainter()

    if seqtype == 'aa':
        return AminoAcidAlignment(alignment), painter
    if seqtype == 'dna' or seqtype == 'rna':
        return DnaAlignment(alignment), DnaPainter()
    elif seqtype == 'codon':
        al = CodonAlignment(alignment)
        al.set_genetic_code(genetic_code)
        return CodonAlignment(alignment), CodonPainter(painter)
    else:
        raise IOError(f'Unknown sequence type: "{seqtype}"')



def output_al_info(alignment):
    for headline, data in alignment.get_basic_info():
        print(headline+':', data)





def guess_seq_type(al):
    '''
    Guess whether alignment al is AA, DNA, coding DNA, or RNA and
    return an appropriate string. RNA is treated as DNA.
    The probabilities are very rough estimates/guesses (and computed in
    log-space).
    '''
    p_aa = _likelihood_of_seq(al, _aa_distr)
    p_dna = _likelihood_of_seq(al, _dna_distr)
    if p_dna < p_aa:
        return 'aa'
    else:
        p_coding = _likelihood_of_codons(al)
        if p_dna > p_coding:
            return 'dna'
        else:
            return 'codon'

def _likelihood_of_seq(al, distr):
    log_p = 0.0
    tiny_prob = min(distr.values()) # For indels etc
    for rec in al:
        for c in rec.seq:
            if c in distr:
                log_p += distr[c]
            else:
                log_p += tiny_prob # Need to penalize the indels, otherwise codons are never chosen
    return log_p

def _likelihood_of_codons(al):
    log_p = 0.0
    tiny_prob = 2*min(_codon_distr.values())
    l = len(str(al[0].seq))
    for rec in al:
        for pos in range(0,l,3):
            codon = str(rec.seq[pos:pos+3])
            if codon in _codon_distr:
                log_p += _codon_distr[codon]
            else:
                log_p += tiny_prob # Need to penalize the lack of proper codon with a small penalty
    return log_p

# Overkill: using amino acid frequencies given for vertebrates, found somewhere on the internet.
_aa_distr = {
    'A': log(0.074),
    'R': log(0.042),
    'N': log(0.044),
    'D': log(0.059),
    'C': log(0.033),
    'E': log(0.058),
    'Q': log(0.037),
    'G': log(0.074),
    'H': log(0.029),
    'I': log(0.038),
    'L': log(0.076),
    'K': log(0.072),
    'M': log(0.018),
    'F': log(0.04),
    'P': log(0.05),
    'S': log(0.081),
    'T': log(0.062),
    'W': log(0.013),
    'Y': log(0.033),
    'V': log(0.066),            # Should be 0.068, but I am giving away 0.1+0.1 to X and *
    'X': log(0.01),
    '*': log(0.01)
    }
_dna_distr = {
    # Nucleotide frequncies are rounded down somewhat, leaving 2% for the other letters. 2/18=0.11111111.
    'A': log(0.30),
    'C': log(0.21),
    'G': log(0.26),
    'T': log(0.22),
    'U': log(0.22),             # It is either DNA or RNA and at this point I don't care.

    # Remaining (?) letters
    'B': log(0.0011),
    'R': log(0.0011),
    'N': log(0.0011),
    'D': log(0.0011),
    'E': log(0.0011),
    'Q': log(0.0011),
    'H': log(0.0011),
    'I': log(0.0011),
    'L': log(0.0011),
    'K': log(0.0011),
    'M': log(0.0011),
    'F': log(0.0011),
    'P': log(0.0011),
    'S': log(0.0011),
    'W': log(0.0011),
    'Y': log(0.0011),
    'V': log(0.0011),
    'X': log(0.0011),
    }

_codon_distr = {
    'TTT': log(0.0169),
    'TTC': log(0.0204),
    'TTA': log(0.0072),
    'TTG': log(0.0126),

    'TCT': log(0.0146),
    'TCC': log(0.0174),
    'TCA': log(0.0117),
    'TCG': log(0.0045),

    'TAT': log(0.012),
    'TAC': log(0.0156),
    'TAA': log(0.0007),         # STOP
    'TAG': log(0.0005),         # STOP

    'TGT': log(0.0099),
    'TGC': log(0.0122),
    'TGA': log(0.0013),         # STOP
    'TGG': log(0.0128),

    'CTT': log(0.0128),
    'CTC': log(0.0194),
    'CTA': log(0.0069),
    'CTG': log(0.0403),

    'CCT': log(0.0173),
    'CCC': log(0.020),
    'CCA': log(0.0167),
    'CCG': log(0.007),

    'CAT': log(0.0104),
    'CAC': log(0.0149),
    'CAA': log(0.0118),
    'CAG': log(0.0346),

    'CGT': log(0.0047),
    'CGC': log(0.0109),
    'CGA': log(0.0063),
    'CGG': log(0.0119),

    'ATT': log(0.0157),
    'ATC': log(0.0214),
    'ATA': log(0.0071),
    'ATG': log(0.0223),

    'ACT': log(0.0128),
    'ACC': log(0.0192),
    'ACA': log(0.0148),
    'ACG': log(0.0062),

    'AAT': log(0.0167),
    'AAC': log(0.0195),
    'AAA': log(0.024),
    'AAG': log(0.0329),

    'AGT': log(0.0119),
    'AGC': log(0.0194),
    'AGA': log(0.0115),
    'AGG': log(0.0114),

    'GTT': log(0.0109),
    'GTC': log(0.0146),
    'GTA': log(0.007),
    'GTG': log(0.0289),

    'GCT': log(0.0186),
    'GCC': log(0.0285),
    'GCA': log(0.016),
    'GCG': log(0.0076),

    'GAT': log(0.0223),
    'GAC': log(0.026),
    'GAA': log(0.029),
    'GAG': log(0.0408),

    'GGT': log(0.0108),
    'GGC': log(0.0228),
    'GGA': log(0.0163),
    'GGG': log(0.0164),
    }



import Bio.Seq
from colorama import init as colorama_init, Fore, Back, Style
import sys

class Painter:
    '''
    Base class for all painters
    '''
    def __init__(self):
        self.restrictions = []

    def set_options(self, args):
        if args.keep_colors_when_redirecting:
            colorama_init(strip=False)
        else:
            colorama_init()
        if args.majority:
            self.restrictions.append(restrict_to_majority)
        if args.no_indels:
            self.restrictions.append(restrict_to_no_indels)
        if args.only_variable:
            self.restrictions.append(restrict_to_variable)
        if args.only_variable_excluding_indels:
            self.restrictions.append(restrict_to_variable_excluding_indels)

    def color_for_bad_data(self):
        return Back.WHITE + Fore.RED, Fore.BLACK + Back.WHITE

    def color_for_stop(self):
        return Back.BLACK + Fore.RED, Fore.BLACK

    def indel_color(self):
        return Back.WHITE, Back.WHITE

    def eol(self):
        return Style.RESET_ALL

    def sol(self):
        return Fore.BLACK + Back.WHITE

    def colorizer(self, c, column):
        if c == '!?':
            before_color, after_color = self.color_for_bad_data()
            return before_color + c + after_color
        elif all(map(lambda r: r(column), self.restrictions)): # True also if the list self.restrictions is empty
            before_color, after_color = self._color_lookup(c)
            colored_item = before_color + c
            if after_color:
                return colored_item + after_color
            else:
                return colored_item
        else:
            before_color, after_color = self.indel_color()
            if after_color:
                return before_color + c + after_color
            else:
                return before_color + c


# Restriction functions
#
# These look at columns, given as a Counter object, and decides whether the column
# should be colored or not. Returns True for 'color it', or False for 'dont color it'.
#
def restrict_to_majority(column):
    maj = column.most_common(1)[0] # Single most common
    if maj[0] != '-' and maj[1] >= 0.5 * len(list(column.elements())):
        return True
    else:
        return False

def restrict_to_no_indels(column):
    return column['-'] == 0

def restrict_to_variable(column):
    return len(column) > 1

def restrict_to_variable_excluding_indels(column):
    if '-' in column:
        return len(column) > 2
    else:
        return len(column) > 1


class AminoAcidPainter(Painter):
    '''
    Put paint of amino acids.
    '''
    def __init__(self):
        super().__init__()

    def _color_lookup(self, c):
        if c in 'AILMFWVCailmfwvc':
            return Back.BLUE, Back.WHITE
        elif c in 'KRkr':
            return Back.RED, Back.WHITE
        elif c in 'EDed':
            return Back.MAGENTA, Back.WHITE
        elif c in 'NQSTnqst':
            return Back.GREEN, Back.WHITE
        elif c in 'Gg':
            return Back.YELLOW, Back.WHITE
        elif c in 'Pp':
            return Back.YELLOW, Back.WHITE
        elif c in 'HYhy':
            return Back.CYAN, Back.WHITE
        elif c in 'Xx':
            return Back.WHITE, Back.WHITE
        elif c in "!?":
            return self.color_for_bad_data()
        elif c in "*":
            return self.color_for_stop()
        elif c in '-_.:':
            return self.indel_color()
        else:
            return Back.WHITE, None


class AminoAcidTaylorPainter(AminoAcidPainter):
    '''
    Put paint to amino acids, an approximation of the "Taylor" style.
    '''
    def __init__(self):
        super().__init__()

    def _color_lookup(self, c):
        if c in 'AILMFVCailMFVC':
            return Back.GREEN, Back.WHITE
        elif c in 'KRHkrh':
            return Back.BLUE, Back.WHITE
        elif c in 'EDSTedst':
            return Back.RED, Back.WHITE
        elif c in 'NQnq':
            return Back.MAGENTA, Back.WHITE
        elif c in 'CGPcgp':
            return Back.YELLOW, Back.WHITE
        elif c in 'Yy':
            return Back.CYAN, Back.WHITE
        elif c in 'X':
            return Back.WHITE, Back.WHITE
        elif c in "!?":
            return self.color_for_bad_data()
        elif c in "*":
            return self.color_for_stop()
        elif c in '-_.:':
            return self.indel_color()
        else:
            return Back.WHITE, None

class AminoAcidHydrophobicity(AminoAcidPainter):
    '''
    Put paint to amino acids, indicating hydrophobicity.
    '''
    def __init__(self):
        super().__init__()

    def _color_lookup(self, c):
        if c in 'AILMFVPGailmfvpg':
            return Back.RED, Back.WHITE
        elif c in 'QNHSTYCWqnhstycw':
            return Back.BLUE, Back.WHITE
        elif c in 'RKDErkde':
            return Back.GREEN, Back.WHITE
        elif c in 'X':
            return Back.WHITE, Back.WHITE
        elif c in "!?":
            return self.color_for_bad_data()
        elif c in "*":
            return self.color_for_stop()
        elif c in '-_.:':
            return self.indel_color()
        else:
            return Back.WHITE, None


class DnaPainter(Painter):
    '''
    Put paint of nucleotides.
    '''
    def __init__(self):
        super().__init__()

    def _color_lookup(self, c):
        if c in 'TUtu':           # Handles RNA too
            return Back.CYAN, Back.WHITE
        elif c in 'Aa':
            return Back.GREEN, Back.WHITE
        elif c in 'Cc':
            return Back.YELLOW , Back.WHITE
        elif c in 'Gg':
            return Back.RED, Back.WHITE
        elif c in '!*':
            return self.color_for_bad_data()
        elif c in '-.:':
            return self.indel_color()
        else:
            return Back.WHITE


class DnaClassPainter(Painter):
    '''
    Put paint of nucleotides.
    '''
    def __init__(self):
        super().__init__()

    def _color_lookup(self, c):
        if c in 'TUtuCcYy':           # Handles RNA too
            return Back.Cyan, Back.White
        elif c in 'AaGgRr':
            return Back.Magenta, Back.White
        elif c in '!*':
            return self.color_for_bad_data()
        elif c in '-.:':
            return self.indel_color()
        else:
            return Back.WHITE



class CodonPainter(Painter):
    '''
    Put paint on codons.
    '''
    def __init__(self, aa_painter):
        super().__init__()
        self.aa_painter = aa_painter

    def colorizer(self, c, column):
        '''
        c is a expected to be a codon, or a single letter.
        Single letters may occur in MACSE alignments, due to frameshifts.
        '''
        try:
            if '!' in c:
                before_color, after_color = self.color_for_bad_data()
                assert after_color # Weird if we don't have a color reset.
                return before_color + c + after_color
            elif all(map(lambda r: r(column), self.restrictions)): # True also if the list self.restrictions is empty
                if len(c) != 3:
                    return Back.WHITE + Fore.RED + Style.BRIGHT + c + Fore.BLACK + Style.NORMAL
                elif c == '---':
                    before_color, after_color = self.indel_color()
                else:
                    aa = Bio.Seq.translate(c)
                    before_color, after_color = self.aa_painter._color_lookup(aa)
                colored_item = before_color + c
                if after_color:
                    return colored_item + after_color
                else:
                    return colored_item
            else:
                before_color, after_color = self.indel_color()
                if after_color:
                    return before_color + c + after_color
                else:
                    return before_color + c

        except Bio.Data.CodonTable.TranslationError:
            return c
        except Exception as e:
            print('Alv warning:', str(e), file=sys.stderr)
            return c                # Temporarily. Adding colors later



import unittest
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from Bio.Align import MultipleSeqAlignment
from alv.alignment import AminoAcidAlignment

class TestAlignment(unittest.TestCase):
    def setUp(self):
        simple_al = MultipleSeqAlignment([
             SeqRecord(Seq("AR"), id="Alpha", annotations={'molecule_type': 'protein'}),
             SeqRecord(Seq("AR"), id="Beta_", annotations={'molecule_type': 'protein'}),
             SeqRecord(Seq("AS"), id="Gamma", annotations={'molecule_type': 'protein'}),
         ])
        self.al = AminoAcidAlignment(simple_al)

    def test_summarize_columns(self):
        self.assertEqual(self.al._summarize_columns()[0]['A'], 3)
        self.assertEqual(self.al._summarize_columns()[1]['R'], 2)
        self.assertEqual(self.al._summarize_columns()[1]['S'], 1)
        self.assertEqual(self.al._summarize_columns()[1]['A'], 0)

    def test_accession_length(self):
        for acc in self.al.accessions():
            self.assertEqual(len(acc), 5)

    def test_basic_info(self):
        info = list(self.al.get_basic_info())
        self.assertEqual(info[0][1], 3)
        self.assertEqual(info[1][1], 2)

    def test_column_conservation(self):
        conservation = self.al.get_column_conservation()
        self.assertEqual(conservation[0], 1.0)
        self.assertEqual(conservation[1], 2/3)

if __name__ == '__main__':
    unittest.main()



from .io import get_alv_objects
from .alignmentterminal import AlignmentTerminal

def view(msa, seqtype='guess', color_scheme='guess', genetic_code=1, width=80, dotted=False):
    '''
    Present a colorised version of a multiple sequence alignment (MSA).

    Options:
    * msa           An alignment as BioPython object (from AlignIO).
    * seqtype       String describing the type of sequence. One of 'aa', 'dna',
                    'rna', 'codon', or 'guess' (which almost always works). 
    * color_scheme  String indicating what coloring scheme to use. Ignore unless
                    you want to use 'taylor' or 'hydrophobicity'.
    * genetic_code  An integer indicating which genetic code to use. In the 
                    range 1 to 33 as of this writing, with 1 being the standard
                    code. See NCBI's list at 
                       https://www.ncbi.nlm.nih.gov/Taxonomy/Utils/wprintgc.cgi
    * width         Integer indicating the desired width of alignment view. 
                    Default: 60 characters.
    * dotted        If True, all characters in the first sequence of the MSA is
                    written and remaining sequences are dotted everwhere except
                    for where they differ from the first sequence.
    '''
    alignment, painter = get_alv_objects(msa, seqtype, color_scheme, genetic_code)
    terminal = AlignmentTerminal(width)
    terminal.output_alignment(alignment, painter, width - 12, dotted)


def glimpse(msa, n_seq=20, seqtype='guess', color_scheme='guess', genetic_code=1, width=60, dotted=False):
    '''
    Works like "view", but extracts a conserved region and outputs a random sample of the sequences or all of them if it is a small alignment.

    Options:
    * msa           An alignment as BioPython object (from AlignIO).
    * n_seq         The number of sequences to output.
    * seqtype       String describing the type of sequence. One of 'aa', 'dna',
                    'rna', 'codon', or 'guess' (which almost always works). 
    * color_scheme  String indicating what coloring scheme to use. Ignore unless
                    you want to use 'taylor' or 'hydrophobicity'.
    * genetic_code  An integer indicating which genetic code to use. In the 
                    range 1 to 33 as of this writing, with 1 being the standard
                    code. See NCBI's list at 
                       https://www.ncbi.nlm.nih.gov/Taxonomy/Utils/wprintgc.cgi
    * width         Integer indicating the desired width of alignment view. 
                    Default: 60 characters.
    * dotted        If True, all characters in the first sequence of the MSA is
                    written and remaining sequences are dotted everwhere except
                    for where they differ from the first sequence.
    '''
    alignment, painter = get_alv_objects(msa, seqtype, color_scheme, genetic_code)
    terminal = AlignmentTerminal(width)
    terminal.height = n_seq + 2
    terminal.output_glimpse(alignment, painter, width, dotted)




import setuptools
import sys


def read_version_string(filename):
    versionstring = None
    with open(filename) as h:
        for line in h:
            try:
                identifier, equalsign, versionstring = line.split()
                if identifier == '__version__' and equalsign == '=':
                    return versionstring.strip("'")
            except:
                continue        # Just read past lines to fitting the pattern
    if versionstring is None:
        return 'unknown_version'

__version__ = read_version_string('alv/version.py')


with open("README.md", "r") as fh:
    at_top = True
    long_description = ''
    for line in fh:
        if at_top and line[:3] == '[![':
            pass                # Skipping the badge-lines in the github README.md
        else:
            at_top = False      # Now starts the "real" README.md
            long_description += line


if sys.version_info.major < 3:
    sys.exit('\n'
             'Sorry, Python 2 is not supported\n'
             'Did you run pip install alv?\n'
             'Try \'pip3 install alv\'')

elif sys.version_info.minor < 5:
    sys.exit('\nSorry, Python < 3.5 is not supported\n')

setuptools.setup(
    name="alv",
    version=__version__,
    author="Lars Arvestad",
    author_email="arve@math.su.se",
    description="A console-based sequence alignment viewer",
    scripts = ['bin/alv'],
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/arvestad/alv",
    test_suite = "tests",
    packages=setuptools.find_packages(),
    python_requires='>=3.5',
    install_requires=[
        'biopython>=1.70',
        'colorama>=0.3.8',
    ],
    classifiers=(
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ),
)
